#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import asyncio
import json
import logging
from asyncio.exceptions import TimeoutError

import websockets
from websockets.exceptions import (
    ConnectionClosedError,
    ConnectionClosedOK,
    InvalidStatusCode,
    WebSocketException,
)

from connect.client import AsyncConnectClient, ConnectClient
from connect.eaas.dataclasses import CapabilitiesPayload, Message, MessageType
from connect.eaas.helpers import (
    get_environment,
    get_extension_class,
    get_extension_type,
)
from connect.eaas.logging import ExtensionLogHandler
from connect.eaas.manager import TasksManager


logger = logging.getLogger(__name__)


class Worker:
    """
    The Worker is responsible to handle the websocket connection
    with the server. It will send the extension capabilities to
    the server and wait for tasks that need to be processed using
    the tasks manager.
    """
    def __init__(self, secure=True):
        self.secure = secure
        env = get_environment()
        self.ws_address = env['ws_address']
        self.api_address = env['api_address']
        self.api_key = env['api_key']
        self.environment_id = env['environment_id']
        self.instance_id = env['instance_id']
        self.headers = (('Authorization', self.api_key),)
        proto = 'wss' if secure else 'ws'
        self.base_ws_url = f'{proto}://{self.ws_address}/public/v1/devops/ws'
        self.run_event = asyncio.Event()
        self.ws = None
        self.extension_class = get_extension_class()
        self.extension_type = get_extension_type(self.extension_class)
        descriptor = self.extension_class.get_descriptor()
        self.capabilities = descriptor['capabilities']
        self.readme_url = descriptor['readme_url']
        self.changelog_url = descriptor['changelog_url']
        self.extension_config = None
        self.logging_api_key = None
        self.main_task = None
        self.tasks_manager = None
        self.paused = False
        self.logging_handler = None
        self.environment_type = None

    async def ensure_connection(self):
        """
        Ensure that a websocket connection is established.
        """
        if self.ws is None or self.ws.closed:
            url = self.get_url()
            self.ws = await websockets.connect(
                url,
                extra_headers=self.headers,
            )
            await (await self.ws.ping())
            logger.info(f'Connected to {url}')

    async def send(self, message):
        """
        Send a message to the websocket server.
        """
        await self.ws.send(json.dumps(message))

    async def receive(self):
        """
        Receive a message from the websocket server.
        """
        try:
            message = await asyncio.wait_for(self.ws.recv(), timeout=1)
            return json.loads(message)
        except TimeoutError:
            pass

    def get_client(self):
        """
        Get an instance of the Connect Openapi Client. If the extension is asyncrhonous
        it returns an instance of the AsyncConnectClient otherwise the ConnectClient.
        """
        client_class = ConnectClient if self.extension_type == 'sync' else AsyncConnectClient
        return client_class(
            self.api_key,
            endpoint=f'https://{self.api_address}/public/v1',
            use_specs=False,
        )

    def get_extension_logger(self, token):
        """
        Returns a logger instance configured with the LogZ.io handler.
        This logger will be used by the extension to send logging records
        to the Logz.io service.
        """
        logger = logging.getLogger('eaas.extension')
        if self.logging_handler is None and token is not None:
            self.logging_handler = ExtensionLogHandler(
                token,
                default_extra_fields={
                    'environment_id': self.environment_id,
                    'instance_id': self.instance_id,
                    'environment_type': self.environment_type,
                },
            )
            logger.addHandler(self.logging_handler)
        return logger

    async def stop_tasks_manager(self):
        if self.tasks_manager:
            logger.debug('shutting down tasks worker....')
            await self.tasks_manager.stop()

    def start_tasks_manager(self):
        logger.info('Starting tasks worker...')
        self.tasks_manager = TasksManager(self)
        self.tasks_manager.start()
        logger.info('Task worker started')

    def ensure_tasks_manager_running(self):
        if not self.paused and (self.tasks_manager is None or not self.tasks_manager.is_running()):
            self.start_tasks_manager()

    def get_url(self):
        return f'{self.base_ws_url}/{self.environment_id}/{self.instance_id}'

    def get_extension(self):
        return self.extension_class(
            self.get_client(),
            self.get_extension_logger(self.logging_api_key),
            self.extension_config,
        )

    async def run(self):  # noqa: CCR001
        """
        Main loop for the websocket connection.
        Once started, this worker will send the capabilities message to
        the websocket server and start a loop to receive messages from the
        websocket server.
        """
        while self.run_event.is_set():
            try:
                await self.ensure_connection()
                message = Message(
                    message_type=MessageType.CAPABILITIES,
                    data=CapabilitiesPayload(
                        self.capabilities,
                        self.readme_url,
                        self.changelog_url,
                    ),
                )
                await self.send(message.to_json())
                while self.run_event.is_set():
                    await self.ensure_connection()
                    self.ensure_tasks_manager_running()
                    message = await self.receive()
                    if not message:
                        continue
                    await self.process_message(message)
            except ConnectionClosedOK:
                break
            except ConnectionClosedError:
                logger.warning(f'Disconnected from: {self.get_url()}, retry in 2s')
                await asyncio.sleep(2)
            except InvalidStatusCode as ic:
                if ic.status_code == 502:
                    logger.warning('Maintenance in progress, try to reconnect in 2s')
                    await asyncio.sleep(2)
                else:
                    logger.warning(f'Received an unexpected status from server: {ic.status_code}')
                    await asyncio.sleep(2)
            except WebSocketException:
                logger.exception('Unexpected websocket exception.')
                await asyncio.sleep(2)

        await self.stop_tasks_manager()
        if self.ws:
            await self.ws.close()

    async def process_message(self, data):
        """
        Process a message received from the websocket server.
        """
        message = Message(**data)
        if message.message_type == MessageType.CONFIGURATION:
            await self.configuration(message.data)
        elif message.message_type == MessageType.TASK:
            await self.tasks_manager.submit_task(message.data)
        elif message.message_type == MessageType.PAUSE:
            await self.pause()
        elif message.message_type == MessageType.RESUME:
            await self.resume()
        elif message.message_type == MessageType.SHUTDOWN:
            await self.shutdown()

    async def configuration(self, data):
        """
        Process the configuration message.
        It will stop the tasks manager so the extension can be
        reconfigured, then restart the tasks manager.
        """
        if data.configuration:
            self.extension_config = data.configuration
        if data.logging_api_key:
            self.logging_api_key = data.logging_api_key
        if data.environment_type:
            self.environment_type = data.environment_type
        if data.log_level:
            logging.getLogger('eaas.extension').setLevel(
                getattr(logging, data.log_level),
            )
        if data.runner_log_level:
            logging.getLogger('connect.eaas').setLevel(
                getattr(logging, data.runner_log_level),
            )
        logger.info('Extension configuration has been updated.')

    async def pause(self):
        """
        Stop the task manager. No task will be consumed
        until a "resume" message is received.
        """
        self.paused = True
        logger.info('Pause task manager operations.')
        await self.stop_tasks_manager()

    async def resume(self):
        """
        Restart the task manager so it will consume tasks once again.
        """
        self.paused = False
        logger.info('Resume task manager operations.')
        self.start_tasks_manager()

    async def shutdown(self):
        """
        Shutdown the extension runner.
        """
        logger.info('Shutdown extension runner.')
        await self.pause()
        self.stop()

    async def start(self):
        logger.info('Starting control worker...')
        self.main_task = asyncio.create_task(self.run())
        logger.info('Control worker started')
        self.run_event.set()
        await self.main_task
        logger.info('Control worker stopped')

    def stop(self):
        logger.info('Stopping control worker.....')
        self.run_event.clear()



import dataclasses
from typing import Any, Optional, Union


class TaskType:
    ASSET_PURCHASE_REQUEST_PROCESSING = 'asset_purchase_request_processing'
    ASSET_CHANGE_REQUEST_PROCESSING = 'asset_change_request_processing'
    ASSET_SUSPEND_REQUEST_PROCESSING = 'asset_suspend_request_processing'
    ASSET_RESUME_REQUEST_PROCESSING = 'asset_resume_request_processing'
    ASSET_CANCEL_REQUEST_PROCESSING = 'asset_cancel_request_processing'
    ASSET_ADJUSTMENT_REQUEST_PROCESSING = 'asset_adjustment_request_processing'
    ASSET_PURCHASE_REQUEST_VALIDATION = 'asset_purchase_request_validation'
    ASSET_CHANGE_REQUEST_VALIDATION = 'asset_change_request_validation'
    PRODUCT_ACTION_EXECUTION = 'product_action_execution'
    PRODUCT_CUSTOM_EVENT_PROCESSING = 'product_custom_event_processing'
    TIER_CONFIG_SETUP_REQUEST_PROCESSING = 'tier_config_setup_request_processing'
    TIER_CONFIG_CHANGE_REQUEST_PROCESSING = 'tier_config_change_request_processing'
    TIER_CONFIG_ADJUSTMENT_REQUEST_PROCESSING = 'tier_config_adjustment_request_processing'
    TIER_CONFIG_SETUP_REQUEST_VALIDATION = 'tier_config_setup_request_validation'
    TIER_CONFIG_CHANGE_REQUEST_VALIDATION = 'tier_config_change_request_validation'


class MessageType:
    CAPABILITIES = 'capabilities'
    CONFIGURATION = 'configuration'
    TASK = 'task'
    PAUSE = 'pause'
    RESUME = 'resume'
    SHUTDOWN = 'shutdown'


class TaskCategory:
    BACKGROUND = 'background'
    INTERACTIVE = 'interactive'


class ResultType:
    SUCCESS = 'success'
    RESCHEDULE = 'reschedule'
    SKIP = 'skip'
    RETRY = 'retry'
    FAIL = 'fail'


@dataclasses.dataclass
class TaskPayload:
    task_id: str
    task_category: str
    task_type: str
    object_id: str
    result: str = None
    data: Any = None
    countdown: int = 0
    output: str = None
    correlation_id: str = None
    reply_to: str = None

    def to_json(self):
        return dataclasses.asdict(self)


@dataclasses.dataclass
class ConfigurationPayload:
    configuration: dict = None
    logging_api_key: str = None
    environment_type: str = None
    log_level: str = None
    runner_log_level: str = None

    def to_json(self):
        return dataclasses.asdict(self)


@dataclasses.dataclass
class CapabilitiesPayload:
    capabilities: dict
    readme_url: str = None
    changelog_url: str = None

    def to_json(self):
        return dataclasses.asdict(self)


@dataclasses.dataclass(init=False)
class Message:
    message_type: str
    data: Optional[Union[CapabilitiesPayload, ConfigurationPayload, TaskPayload]] = None

    def __init__(self, message_type=None, data=None):
        self.message_type = message_type
        if isinstance(data, dict):
            if self.message_type == MessageType.CONFIGURATION:
                self.data = ConfigurationPayload(**data)
            elif self.message_type == MessageType.TASK:
                self.data = TaskPayload(**data)
            elif self.message_type == MessageType.CAPABILITIES:
                self.data = CapabilitiesPayload(**data)
        else:
            self.data = data

    def to_json(self):
        payload = {'message_type': self.message_type}
        if self.data:
            payload['data'] = dataclasses.asdict(self.data)
        return payload



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import asyncio
from asyncio.futures import Future
import inspect
import logging
from concurrent.futures import ThreadPoolExecutor

from connect.client import AsyncConnectClient, ClientError
from connect.eaas.constants import (
    ASSET_REQUEST_TASK_TYPES,
    BACKGROUND_TASK_MAX_EXECUTION_TIME,
    INTERACTIVE_TASK_MAX_EXECUTION_TIME,
    RESULT_SENDER_MAX_RETRIES,
    TASK_TYPE_EXT_METHOD_MAP,
    TIER_CONFIG_REQUEST_TASK_TYPES,
)
from connect.eaas.dataclasses import (
    Message,
    MessageType,
    ResultType,
    TaskCategory,
    TaskPayload,
)
from connect.eaas.extension import ProcessingResponse


logger = logging.getLogger(__name__)


class TasksManager:
    """
    The TasksManager is responsible for dispatching tasks to the
    corresponding methods of the extension.
    When a task is submitted, the TasksManager executes the method
    of the extension in a worker thread if the extension method is
    synchronous otherwise it schedules its execution as an asyncio task.
    Once submitted, an asyncio task will be started that will wait for
    the future returned by the thread pool to be completed.
    Once completed, it will push the result to a queue from which, the
    result_sender task will pick it and send the result to the backend.
    """
    def __init__(self, worker):
        self.worker = worker
        self.background_executor = ThreadPoolExecutor()
        self.interactive_executor = ThreadPoolExecutor()
        self.run_event = asyncio.Event()
        self.loop = asyncio.get_running_loop()
        self.client = AsyncConnectClient(
            self.worker.api_key,
            endpoint=f'https://{self.worker.api_address}/public/v1',
            use_specs=False,
        )
        self.results_queue = asyncio.Queue()
        self.running_tasks = 0
        self.main_task = None
        self.sender_task = None

    def start(self):
        logger.info('Starting tasks manager...')
        self.main_task = asyncio.create_task(self.result_sender())
        logger.info('Tasks manager started')
        self.run_event.set()

    async def stop(self):
        logger.info('Stopping tasks manager...')
        self.run_event.clear()
        await self.main_task
        logger.info('Tasks manager stopped')

    def is_running(self):
        return self.run_event.is_set()

    async def submit_task(self, data):
        """
        Submit a new task to a worker thread if the extension handler
        method is synchronous, otherwise schedules its execution as an
        asyncio task. It also retrieves the task's related object through
        the Connect public API and pass it to the handler method.
        """
        logger.info(
            f'submit new task: id={data.task_id}, '
            f'type={data.task_type}, object={data.object_id}',
        )
        object_id = data.object_id
        task_type = data.task_type
        method_name = TASK_TYPE_EXT_METHOD_MAP[task_type]
        extension = self.worker.get_extension()
        method = getattr(extension, method_name)
        logger.debug(f'invoke {method_name}')
        self.running_tasks += 1
        request = None
        if data.task_category == TaskCategory.BACKGROUND:
            try:
                request = await self.get_request(object_id, task_type)
            except ClientError as e:
                logger.warning(f'Cannot retrieve object {data.object_id} for task {data.task_id}')
                self.send_exception_response(data, e)
                return
            request_status = request.get('status')
            if request_status not in self.worker.capabilities[task_type]:
                logger.debug('Send skip response since request status is not supported.')
                self.send_skip_response(
                    data,
                    f'The status {request_status} is not supported by the extension.',
                )
                return
        else:
            request = data.data

        if inspect.iscoroutinefunction(method):
            future = asyncio.create_task(method(request))
        else:
            future = self.loop.run_in_executor(
                self.get_executor(task_type),
                method,
                request,
            )
        logger.info(f'Task {data.task_id} has been submitted.')
        logger.debug(f'Current running task: {self.running_tasks}')
        asyncio.create_task(self.enqueue_result(data, future))

    def get_executor(self, task_category):
        """
        Return a thread pool executor based on the task category.
        """
        return (
            self.background_executor
            if task_category == TaskCategory.BACKGROUND
            else self.interactive_executor
        )

    async def get_request(self, object_id, task_type):
        """
        Get the request object through the Connect public API
        related to the task that need processing.
        """
        if task_type in ASSET_REQUEST_TASK_TYPES:
            logger.debug(f'get asset request {object_id}')
            return await self.client.requests[object_id].get()
        if task_type in TIER_CONFIG_REQUEST_TASK_TYPES:
            logger.debug(f'get TC request {object_id}')
            return await self.client('tier').config_requests[object_id].get()

    async def result_sender(self):  # noqa: CCR001
        """
        This coroutine is responsible of dequeueing results from
        the results queue and send it to the EaaS backend.
        """
        while True:
            if self.results_queue.empty():
                if not self.run_event.is_set() and self.running_tasks == 0:
                    logger.info('Worker exiting and no more running tasks: exit!')
                    return
                await asyncio.sleep(.1)
                continue
            result = await self.results_queue.get()
            logger.debug(f'got result from queue: {result}')
            retries = 0
            while retries < RESULT_SENDER_MAX_RETRIES:
                try:
                    message = Message(
                        message_type=MessageType.TASK,
                        data=result,
                    )
                    await self.worker.send(message.to_json())
                    logger.info(f'Result for task {result.task_id} has been sent.')
                    break
                except Exception:
                    retries += 1
                    await asyncio.sleep(.1)
            else:
                logger.warning(
                    f'max retries exceeded for sending results of task {result.task_id}',
                )

            if not self.run_event.is_set():
                logger.debug(
                    f'Current processing status: running={self.running_tasks} '
                    f'results={self.results_queue.qsize()}',
                )

    async def enqueue_result(self, task_data, future):
        """
        Build a result object for a task and put it in the
        result queue.
        """
        if task_data.task_category == TaskCategory.BACKGROUND:
            await self.results_queue.put(
                await self.build_bg_response(task_data, future),
            )
        else:
            await self.results_queue.put(
                await self.build_interactive_response(task_data, future),
            )
        self.running_tasks -= 1
        logger.debug(f'enqueue results for sender, running tasks: {self.running_tasks}')

    def send_exception_response(self, data, e):
        future = Future()
        future.set_exception(e)
        asyncio.create_task(self.enqueue_result(data, future))

    def send_skip_response(self, data, output):
        future = Future()
        future.set_result(ProcessingResponse.skip(output))
        asyncio.create_task(self.enqueue_result(data, future))

    async def build_bg_response(self, task_data, future):
        """
        Wait for a background task to be completed and than uild the task result message.
        """
        result_message = TaskPayload(**task_data.to_json())
        result = None
        try:
            result = await asyncio.wait_for(future, timeout=BACKGROUND_TASK_MAX_EXECUTION_TIME)
        except Exception as e:
            logger.warning(f'Got exception during execution of task {task_data.task_id}: {e}')
            result_message.result = ResultType.RETRY
            result_message.output = str(e)
            return result_message
        logger.debug(f'result: {result}')
        result_message.result = result.status

        if result.status == ResultType.SKIP:
            result_message.output = result.output

        if result.status == ResultType.RESCHEDULE:
            result_message.countdown = result.countdown
        return result_message

    async def build_interactive_response(self, task_data, future):
        """
        Wait for an interactive task to be completed and than uild the task result message.
        """
        result = None
        result_message = TaskPayload(**task_data.to_json())
        try:
            result = await asyncio.wait_for(future, timeout=INTERACTIVE_TASK_MAX_EXECUTION_TIME)
        except Exception as e:
            logger.warning(f'Got exception during execution of task {task_data.task_id}: {e}')
            result_message.result = ResultType.FAIL
            result_message.output = str(e)
            return result_message

        result_message.result = result.status
        result_message.data = result.data
        return result_message



from logzio.handler import LogzioHandler


class ExtensionLogHandler(LogzioHandler):
    def __init__(self, *args, **kwargs):
        self.default_extra_fields = kwargs.pop('default_extra_fields')
        super().__init__(*args, **kwargs)

    def extra_fields(self, message):
        extra_fields = super().extra_fields(message)
        extra_fields.update(self.default_extra_fields)
        return extra_fields
